import ast
import asyncio
import functools
import itertools
import logging
import os
import re
import string
from copy import deepcopy
from typing import List, Callable, Dict, Optional, Any, Collection

import pandas as pd
import tiktoken

logger = logging.getLogger("AutoRAG")


def fetch_contents(corpus_data: pd.DataFrame, ids: List[List[str]],
                   column_name: str = 'contents') -> List[List[Any]]:
    flat_ids = itertools.chain.from_iterable(ids)
    contents = list(map(lambda x: corpus_data.loc[lambda row: row['doc_id'] == x][column_name].values[0], flat_ids))

    result = []
    idx = 0
    for sublist in ids:
        result.append(contents[idx:idx + len(sublist)])
        idx += len(sublist)

    return result


def result_to_dataframe(column_names: List[str]):
    """
    Decorator for converting results to pd.DataFrame.
    """

    def decorator_result_to_dataframe(func: Callable):
        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> pd.DataFrame:
            results = func(*args, **kwargs)
            if len(column_names) == 1:
                df_input = {column_names[0]: results}
            else:
                df_input = {column_name: result for result, column_name in zip(results, column_names)}
            result_df = pd.DataFrame(df_input)
            return result_df

        return wrapper

    return decorator_result_to_dataframe


def load_summary_file(summary_path: str,
                      dict_columns: Optional[List[str]] = None) -> pd.DataFrame:
    """
    Load summary file from summary_path.

    :param summary_path: The path of the summary file.
    :param dict_columns: The columns that are dictionary type.
        You must fill this parameter if you want to load summary file properly.
        Default is ['module_params'].
    :return: The summary dataframe.
    """
    if not os.path.exists(summary_path):
        raise ValueError(f"summary.csv does not exist in {summary_path}.")
    summary_df = pd.read_csv(summary_path)
    if dict_columns is None:
        dict_columns = ['module_params']

    if any([col not in summary_df.columns for col in dict_columns]):
        raise ValueError(f"{dict_columns} must be in summary_df.columns.")

    def convert_dict(elem):
        return ast.literal_eval(elem)

    summary_df[dict_columns] = summary_df[dict_columns].applymap(convert_dict)
    return summary_df


def make_combinations(target_dict: Dict[str, Any]) -> List[Dict[str, Any]]:
    """
    Make combinations from target_dict.
    The target_dict key value must be a string,
    and the value can be list of values or single value.
    If generates all combinations of values from target_dict,
    which means generated dictionaries that contain only one value for each key,
    and all dictionaries will be different from each other.

    :param target_dict: The target dictionary.
    :return: The list of generated dictionaries.
    """
    dict_with_lists = dict(map(lambda x: (x[0], x[1] if isinstance(x[1], list) else [x[1]]),
                               target_dict.items()))

    def delete_duplicate(x):
        def is_hashable(obj):
            try:
                hash(obj)
                return True
            except TypeError:
                return False

        if any([not is_hashable(elem) for elem in x]):
            # TODO: add duplication check for unhashable objects
            return x
        else:
            return list(set(x))

    dict_with_lists = dict(map(lambda x: (x[0], delete_duplicate(x[1])), dict_with_lists.items()))
    combination = list(itertools.product(*dict_with_lists.values()))
    combination_dicts = [dict(zip(dict_with_lists.keys(), combo)) for combo in combination]
    return combination_dicts


def explode(index_values: Collection[Any], explode_values: Collection[Collection[Any]]):
    """
    Explode index_values and explode_values.
    The index_values and explode_values must have the same length.
    It will flatten explode_values and keep index_values as a pair.

    :param index_values: The index values.
    :param explode_values: The exploded values.
    :return: Tuple of exploded index_values and exploded explode_values.
    """
    assert len(index_values) == len(explode_values), "Index values and explode values must have same length"
    df = pd.DataFrame({
        'index_values': index_values,
        'explode_values': explode_values
    })
    df = df.explode('explode_values')
    return df['index_values'].tolist(), df['explode_values'].tolist()


def replace_value_in_dict(target_dict: Dict, key: str,
                          replace_value: Any) -> Dict:
    """
    Replace the value of the certain key in target_dict.
    If there is not targeted key in target_dict, it will return target_dict.

    :param target_dict: The target dictionary.
    :param key: The key to replace.
    :param replace_value: The value to replace.
    :return: The replaced dictionary.
    """
    replaced_dict = deepcopy(target_dict)
    if key not in replaced_dict:
        return replaced_dict
    replaced_dict[key] = replace_value
    return replaced_dict


def normalize_string(s: str) -> str:
    """
    Taken from the official evaluation script for v1.1 of the SQuAD dataset.
    Lower text and remove punctuation, articles and extra whitespace.
    """

    def remove_articles(text):
        return re.sub(r"\b(a|an|the)\b", " ", text)

    def white_space_fix(text):
        return " ".join(text.split())

    def remove_punc(text):
        exclude = set(string.punctuation)
        return "".join(ch for ch in text if ch not in exclude)

    def lower(text):
        return text.lower()

    return white_space_fix(remove_articles(remove_punc(lower(s))))


def convert_string_to_tuple_in_dict(d):
    """Recursively converts strings that start with '(' and end with ')' to tuples in a dictionary."""
    for key, value in d.items():
        # If the value is a dictionary, recurse
        if isinstance(value, dict):
            convert_string_to_tuple_in_dict(value)
        # If the value is a list, iterate through its elements
        elif isinstance(value, list):
            for i, item in enumerate(value):
                # If an item in the list is a dictionary, recurse
                if isinstance(item, dict):
                    convert_string_to_tuple_in_dict(item)
                # If an item in the list is a string matching the criteria, convert it to a tuple
                elif isinstance(item, str) and item.startswith('(') and item.endswith(')'):
                    value[i] = ast.literal_eval(item)
        # If the value is a string matching the criteria, convert it to a tuple
        elif isinstance(value, str) and value.startswith('(') and value.endswith(')'):
            d[key] = ast.literal_eval(value)

    return d


def convert_env_in_dict(d: Dict):
    """
    Recursively converts environment variable string in a dictionary to actual environment variable.

    :param d: The dictionary to convert.
    :return: The converted dictionary.
    """
    env_pattern = re.compile(r".*?\${(.*?)}.*?")

    def convert_env(val: str):
        matches = env_pattern.findall(val)
        for match in matches:
            val = val.replace(f"${{{match}}}", os.environ.get(match, ""))
        return val

    for key, value in d.items():
        if isinstance(value, dict):
            convert_env_in_dict(value)
        elif isinstance(value, list):
            for i, item in enumerate(value):
                if isinstance(item, dict):
                    convert_env_in_dict(item)
                elif isinstance(item, str):
                    value[i] = convert_env(item)
        elif isinstance(value, str):
            d[key] = convert_env(value)
    return d


async def process_batch(tasks, batch_size: int = 64) -> List[Any]:
    """
    Processes tasks in batches asynchronously.

    :param tasks: A list of no-argument functions or coroutines to be executed.
    :param batch_size: The number of tasks to process in a single batch.
        Default is 64.
    :return: A list of results from the processed tasks.
    """
    results = []

    for i in range(0, len(tasks), batch_size):
        batch = tasks[i:i + batch_size]
        batch_results = await asyncio.gather(*batch)
        results.extend(batch_results)

    return results


def save_parquet_safe(df: pd.DataFrame, filepath: str,
                      upsert: bool = False):
    output_file_dir = os.path.dirname(filepath)
    if not os.path.isdir(output_file_dir):
        raise NotADirectoryError(f"directory {output_file_dir} not found.")
    if not filepath.endswith("parquet"):
        raise NameError(f'file path: {filepath}  filename extension need to be ".parquet"')
    if os.path.exists(filepath) and not upsert:
        raise FileExistsError(f"file {filepath} already exists."
                              "Set upsert True if you want to overwrite the file.")

    df.to_parquet(filepath, index=False)


def openai_truncate_by_token(texts: List[str], token_limit: int,
                             model_name: str):
    tokenizer = tiktoken.encoding_for_model(model_name)

    def truncate_text(text: str, limit: int, tokenizer):
        tokens = tokenizer.encode(text)
        if len(tokens) <= limit:
            return text
        truncated_text = tokenizer.decode(tokens[:limit])
        return truncated_text

    return list(map(lambda x: truncate_text(x, token_limit, tokenizer), texts))


def reconstruct_list(flat_list: List[Any], lengths: List[int]) -> List[List[Any]]:
    result = []
    start = 0
    for length in lengths:
        result.append(flat_list[start:start + length])
        start += length
    return result



import asyncio
from typing import List, Tuple

import torch
import torch.nn.functional as F

from autorag.nodes.passagereranker.base import passage_reranker_node
from autorag.nodes.passagereranker.tart.modeling_enc_t5 import EncT5ForSequenceClassification
from autorag.nodes.passagereranker.tart.tokenization_enc_t5 import EncT5Tokenizer


@passage_reranker_node
def tart(queries: List[str], contents_list: List[List[str]],
         scores_list: List[List[float]], ids_list: List[List[str]],
         top_k: int, instruction: str = "Find passage to answer given question") \
        -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank a list of contents based on their relevance to a query using Tart.
    TART is a reranker based on TART (https://github.com/facebookresearch/tart).
    You can rerank the passages with the instruction using TARTReranker.
    The default model is facebook/tart-full-flan-t5-xl.

    :param queries: The list of queries to use for reranking
    :param contents_list: The list of lists of contents to rerank
    :param scores_list: The list of lists of scores retrieved from the initial ranking
    :param ids_list: The list of lists of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param instruction: The instruction for reranking.
        Note: default instruction is "Find passage to answer given question"
            The default instruction from the TART paper is being used.
            If you want to use a different instruction, you can change the instruction through this parameter
    :return: tuple of lists containing the reranked contents, ids, and scores
    """
    model_name = "facebook/tart-full-flan-t5-xl"
    model = EncT5ForSequenceClassification.from_pretrained(model_name)
    tokenizer = EncT5Tokenizer.from_pretrained(model_name)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model = model.to(device)
    # Run async tart_rerank_pure function
    tasks = [tart_pure(query, contents, scores, ids, top_k, model, tokenizer, instruction, device) \
             for query, contents, scores, ids in zip(queries, contents_list, scores_list, ids_list)]
    loop = asyncio.get_event_loop()
    results = loop.run_until_complete(asyncio.gather(*tasks))
    content_result = list(map(lambda x: x[0], results))
    id_result = list(map(lambda x: x[1], results))
    score_result = list(map(lambda x: x[2], results))

    del model
    del tokenizer
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    return content_result, id_result, score_result


async def tart_pure(query: str, contents: List[str], scores: List[float],
                    ids: List[str], top_k: int, model, tokenizer, instruction: str, device) \
        -> Tuple[List[str], List[str], List[float]]:
    """
    Rerank a list of contents based on their relevance to a query using Tart.

    :param query: The query to use for reranking
    :param contents: The list of contents to rerank
    :param scores: The list of scores retrieved from the initial ranking
    :param ids: The list of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param model: The Tart model to use for reranking
    :param tokenizer: The tokenizer to use for the model
    :param instruction: The instruction for reranking.
    :param device: The device to run the model on (GPU if available, otherwise CPU)
    :return: tuple of lists containing the reranked contents, ids, and scores
    """

    instruction_queries: List[str] = ['{0} [SEP] {1}'.format(instruction, query) for _ in range(len(contents))]
    features = tokenizer(instruction_queries, contents, padding=True, truncation=True, return_tensors="pt")
    features = features.to(device)

    with torch.no_grad():
        scores = model(**features).logits
        normalized_scores = [float(score[1]) for score in F.softmax(scores, dim=1)]

    contents_ids_scores = list(zip(contents, ids, normalized_scores))

    sorted_contents_ids_scores = sorted(contents_ids_scores, key=lambda x: x[2], reverse=True)

    # crop with top_k
    if len(contents) < top_k:
        top_k = len(contents)
    sorted_contents_ids_scores = sorted_contents_ids_scores[:top_k]

    content_result, id_result, score_result = zip(*sorted_contents_ids_scores)

    return list(content_result), list(id_result), list(score_result)



from collections import defaultdict
from typing import List, Tuple

import pandas as pd

from autorag.nodes.retrieval.base import retrieval_node


@retrieval_node
def hybrid_rsf(ids: Tuple, scores: Tuple, top_k: int,
               weights: Tuple = (0.5, 0.5)) -> Tuple[List[str], List[float]]:
    """
    Hybrid RSF function.
    RSF (Relative Score Fusion) is a method to combine multiple retrieval results based on their relative scores.
    It is uniquer than other retrieval modules, because it does not really execute retrieval,
    but just fuse the results of other retrieval functions.
    So you have to run more than two retrieval modules before running this function.
    And collect ids and scores result from each retrieval module.
    Make it as tuple and input it to this function.

    :param ids: The tuple of ids that you want to fuse.
        The length of this must be the same as the length of scores.
    :param scores: The retrieve scores that you want to fuse.
        The length of this must be the same as the length of ids.
    :param top_k: The number of passages to be retrieved.
    :param weights: Weight for each retrieval result.
        Default is (0.5, 0.5).
        You must set its length as the same as the length of ids and scores.
        Plus, the sum of the weights must be 1.
    :return: The tuple of ids and fused scores that fused by RSF.
    """
    assert len(ids) == len(scores), "The length of ids and scores must be the same."
    assert len(ids) > 1, "You must input more than one retrieval results."
    assert top_k > 0, "top_k must be greater than 0."
    assert sum(weights) == 1, "The sum of weights must be 1."

    # Initialize DataFrame for ids and scores
    id_df = pd.DataFrame({f'id_{i}': id_list for i, id_list in enumerate(ids)})
    score_df = pd.DataFrame({f'score_{i}': score_list for i, score_list in enumerate(scores)})
    df = pd.concat([id_df, score_df], axis=1)

    # Apply relative score fusion
    def rsf_pure_apply(row):
        ids_tuple = tuple(row[[f'id_{i}' for i in range(len(ids))]].values)
        scores_tuple = tuple(row[[f'score_{i}' for i in range(len(scores))]].values)
        return pd.Series(rsf_pure(ids_tuple, scores_tuple, top_k, weights))

    df[['rsf_id', 'rsf_score']] = df.apply(rsf_pure_apply, axis=1)
    return df['rsf_id'].tolist(), df['rsf_score'].tolist()


def rsf_pure(ids: Tuple, scores: Tuple, top_k: int,
             weights: Tuple = (0.5, 0.5)) -> Tuple[List[str], List[float]]:
    min_max_scores = rsf_min_max_scores(scores)
    top_ids, top_scores = score_fusion(ids, scores, top_k, min_max_scores, weights)
    return top_ids, top_scores


def rsf_min_max_scores(scores: Tuple):
    min_max_scores = {}
    for i, score_list in enumerate(scores):
        if not score_list:
            min_max_scores[i] = (0.0, 0.0)
            continue
        min_score = min(score_list)
        max_score = max(score_list)
        min_max_scores[i] = (min_score, max_score)
    return min_max_scores


def score_fusion(ids: Tuple, scores: Tuple, top_k: int, min_max_scores,
                 weights: Tuple = (0.5, 0.5)):
    scaled_scores = []
    for i, (score_list, weight) in enumerate(zip(scores, weights)):
        scaled_score_list = []
        min_score, max_score = min_max_scores[i]
        for score in score_list:
            if max_score == min_score:
                scaled_score = 1.0 if max_score > 0 else 0.0
            else:
                scaled_score = (score - min_score) / (max_score - min_score)
            scaled_score *= weight
            scaled_score_list.append(scaled_score)
        scaled_scores.append(scaled_score_list)

    # Fuse scores
    fused_scores_dict = defaultdict(float)
    for score_list, id_list in zip(scaled_scores, ids):
        for score, doc_id in zip(score_list, id_list):
            fused_scores_dict[doc_id] += score

    # Sort by fused scores and select top_k
    sorted_items = sorted(fused_scores_dict.items(), key=lambda item: item[1], reverse=True)[:top_k]
    top_ids, top_scores = zip(*sorted_items) if sorted_items else ([], [])

    return list(top_ids), list(top_scores)



import functools
import logging
from pathlib import Path
from typing import List, Union

import pandas as pd

from autorag import generator_models
from autorag.utils import result_to_dataframe, validate_qa_dataset

logger = logging.getLogger("AutoRAG")


def query_expansion_node(func):
    @functools.wraps(func)
    @result_to_dataframe(["queries"])
    def wrapper(
            project_dir: Union[str, Path],
            previous_result: pd.DataFrame,
            *args, **kwargs) -> List[List[str]]:
        validate_qa_dataset(previous_result)

        # find queries columns
        assert "query" in previous_result.columns, "previous_result must have query column."
        queries = previous_result["query"].tolist()

        if func.__name__ == "pass_query_expansion":
            return func(queries=queries)

        # set module parameters
        llm_str = kwargs.pop("llm")

        # pop prompt from kwargs
        if "prompt" in kwargs.keys():
            prompt = kwargs.pop("prompt")
        else:
            prompt = ""

        # pop batch from kwargs
        if "batch" in kwargs.keys():
            batch = kwargs.pop("batch")
        else:
            batch = 16

        # set llm model for query expansion
        if llm_str in generator_models:
            llm = generator_models[llm_str](**kwargs)
        else:
            logger.error(f"llm_str {llm_str} does not exist.")
            raise KeyError(f"llm_str {llm_str} does not exist.")

        # run query expansion function
        expanded_queries = func(queries=queries, llm=llm, prompt=prompt, batch=batch)
        del llm
        return expanded_queries

    return wrapper



import logging
import logging.config
import os
import sys

import transformers
from llama_index.embeddings.huggingface import HuggingFaceEmbedding
from llama_index.embeddings.openai import OpenAIEmbedding
from llama_index.embeddings.openai import OpenAIEmbeddingModelType
from llama_index.llms.huggingface import HuggingFaceLLM
from llama_index.llms.openai import OpenAI
from llama_index.llms.openai_like import OpenAILike
from rich.logging import RichHandler
from swifter import set_defaults

version_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'VERSION')

with open(version_path, 'r') as f:
    __version__ = f.read().strip()

set_defaults(allow_dask_on_strings=True)


class LazyInit:
    def __init__(self, factory, *args, **kwargs):
        self._factory = factory
        self._args = args
        self._kwargs = kwargs
        self._instance = None

    def __call__(self):
        if self._instance is None:
            self._instance = self._factory(*self._args, **self._kwargs)
        return self._instance

    def __getattr__(self, name):
        if self._instance is None:
            self._instance = self._factory(*self._args, **self._kwargs)
        return getattr(self._instance, name)


embedding_models = {
    'openai': LazyInit(OpenAIEmbedding),  # default model is OpenAIEmbeddingModelType.TEXT_EMBED_ADA_002
    'openai_embed_3_large': LazyInit(OpenAIEmbedding, model_name=OpenAIEmbeddingModelType.TEXT_EMBED_3_LARGE),
    'openai_embed_3_small': LazyInit(OpenAIEmbedding, model_name=OpenAIEmbeddingModelType.TEXT_EMBED_3_SMALL),
    # you can use your own model in this way.
    'huggingface_baai_bge_small': LazyInit(HuggingFaceEmbedding, model_name="BAAI/bge-small-en-v1.5"),
    'huggingface_cointegrated_rubert_tiny2': LazyInit(HuggingFaceEmbedding, model_name="cointegrated/rubert-tiny2"),
    'huggingface_all_mpnet_base_v2': LazyInit(HuggingFaceEmbedding,
                                              model_name="sentence-transformers/all-mpnet-base-v2",
                                              max_length=512, )
}

generator_models = {
    'openai': OpenAI,
    'huggingfacellm': HuggingFaceLLM,
    'openailike': OpenAILike,
}

rich_format = "[%(filename)s:%(lineno)s] >> %(message)s"
logging.basicConfig(
    level="INFO",
    format=rich_format,
    handlers=[RichHandler(rich_tracebacks=True)]
)
logger = logging.getLogger("AutoRAG")


def handle_exception(exc_type, exc_value, exc_traceback):
    logger = logging.getLogger("AutoRAG")
    logger.error("Unexpected exception",
                 exc_info=(exc_type, exc_value, exc_traceback))


sys.excepthook = handle_exception

transformers.logging.set_verbosity_error()



import os
import pathlib
import tempfile
from datetime import datetime
from uuid import uuid4

import pandas as pd
import pytest

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent.parent.parent
project_dir = os.path.join(root_dir, "resources", "sample_project")
qa_data = pd.read_parquet(os.path.join(project_dir, "data", "qa.parquet"))
corpus_data = pd.read_parquet(os.path.join(project_dir, "data", "corpus.parquet"))
previous_result = qa_data.sample(2)

queries_example = ["What is the capital of France?",
                   "How many members are in Newjeans?"]
contents_example = [["NomaDamas is Great Team", "Paris is the capital of France.", "havertz is suck at soccer"],
                    ["i am hungry", "LA is a country in the United States.", "Newjeans has 5 members."]]
ko_queries_example = ["프랑스의 수도는 어디인가요?",
                      "뉴진스의 멤버는 몇 명인가요?"]
ko_contents_example = [["마커AI는 멋진 회사입니다.", "프랑스의 수도는 파리 입니다.", "아스날은 축구를 못합니다."],
                       ["배고파요", "LA는 미국의 도시입니다.", "뉴진스의 멤버는 5명 입니다."]]
time_list = [[datetime(2021, 1, 1), datetime(2021, 1, 20), datetime(2021, 1, 3)],
             [datetime(2021, 3, 1), datetime(2022, 3, 1), datetime(2030, 1, 1)]]
ids_example = [[str(uuid4()) for _ in range(len(contents_example[0]))],
               [str(uuid4()) for _ in range(len(contents_example[1]))]]
scores_example = [[0.1, 0.8, 0.1], [0.1, 0.2, 0.7]]
f1_example = [0.4, 0.4]
recall_example = [1.0, 1.0]

previous_result['query'] = queries_example
previous_result['retrieved_contents'] = contents_example
previous_result['retrieved_ids'] = ids_example
previous_result['retrieve_scores'] = scores_example
previous_result['retrieval_f1'] = f1_example
previous_result['retrieval_recall'] = recall_example

ko_previous_result = previous_result.copy(deep=True)
ko_previous_result['query'] = ko_queries_example
ko_previous_result['retrieved_contents'] = ko_contents_example


def base_reranker_test(contents, ids, scores, top_k, use_ko=False):
    assert len(contents) == len(ids) == len(scores) == 2
    assert len(contents[0]) == len(ids[0]) == len(scores[0]) == top_k
    for content_list, id_list, score_list in zip(contents, ids, scores):
        assert isinstance(content_list, list)
        assert isinstance(id_list, list)
        assert isinstance(score_list, list)
        for content, _id, score in zip(content_list, id_list, score_list):
            assert isinstance(content, str)
            assert isinstance(_id, str)
            assert isinstance(score, float)
        for i in range(1, len(score_list)):
            assert score_list[i - 1] >= score_list[i]
    if use_ko is True:
        assert contents[0][0] == "프랑스의 수도는 파리 입니다."
        assert ids[0][0] in ids_example[0][1]
        assert contents[1][0] == "뉴진스의 멤버는 5명 입니다."
        assert ids[1][0] in ids_example[1][2]
    else:
        assert contents[0][0] == "Paris is the capital of France."
        assert ids[0][0] in ids_example[0][1]
        assert contents[1][0] == "Newjeans has 5 members."
        assert ids[1][0] in ids_example[1][2]


def base_reranker_node_test(result_df, top_k, use_ko=False):
    contents = result_df["retrieved_contents"].tolist()
    ids = result_df["retrieved_ids"].tolist()
    scores = result_df["retrieve_scores"].tolist()
    base_reranker_test(contents, ids, scores, top_k, use_ko)


@pytest.fixture
def project_dir_with_corpus():
    with tempfile.TemporaryDirectory() as temp_dir:
        data_dir = os.path.join(temp_dir, "data")
        os.makedirs(data_dir, exist_ok=True)
        qa_data.to_parquet(os.path.join(data_dir, "qa.parquet"), index=False)
        new_rows = pd.DataFrame({
            'doc_id': ids_example[0] + ids_example[1],
            'contents': contents_example[0] + contents_example[1],
            'metadata': list(map(lambda x: {'last_modified_datetime': x}, time_list[0])) + list(
                map(lambda x: {'last_modified_datetime': x}, time_list[1]))
        })
        new_corpus = pd.concat([corpus_data, new_rows], ignore_index=True, axis=0)
        new_corpus.to_parquet(os.path.join(data_dir, "corpus.parquet"), index=False)
        yield temp_dir



import pandas as pd

prompt = "Answer this question: {query} \n\n {retrieved_contents}"
queries = ["What is the capital of Japan?", "What is the capital of China?"]
retrieved_contents = [
    ["Tokyo is the capital of Japan.", "Tokyo, the capital of Japan, is a huge metropolitan city."],
    ["Beijing is the capital of China.", "Beijing, the capital of China, is a huge metropolitan city."]]
retrieve_scores = [[0.9, 0.8], [0.9, 0.8]]
previous_result = pd.DataFrame({
    "query": queries,
    "retrieved_contents": retrieved_contents,
    "retrieve_scores": retrieve_scores
})



import os
import pathlib
from uuid import uuid4

import pandas as pd

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent.parent.parent
project_dir = os.path.join(root_dir, "resources", "sample_project")
qa_data = pd.read_parquet(os.path.join(project_dir, "data", "qa.parquet"))
corpus_data = pd.read_parquet(os.path.join(project_dir, "data", "corpus.parquet"))
previous_result = qa_data.sample(2)

queries_example = ["What is the capital of France?",
                   "How many members are in Newjeans?"]
contents_example = [["NomaDamas is Great Team", "Paris is the capital of France.", "havertz is suck at soccer",
                     "Paris is one of the capital from France. Isn't it?"],
                    ["i am hungry", "LA is a country in the United States.", "Newjeans has 5 members.",
                     "Danielle is one of the members of Newjeans."]]
ids_example = [[str(uuid4()) for _ in range(len(contents_example[0]))],
               [str(uuid4()) for _ in range(len(contents_example[1]))]]
scores_example = [[0.1, 0.8, 0.1, 0.5], [0.1, 0.2, 0.7, 0.3]]
f1_example = [0.4, 1.0]
recall_example = [1.0, 0.3]

previous_result['query'] = queries_example
previous_result['retrieved_contents'] = contents_example
previous_result['retrieved_ids'] = ids_example
previous_result['retrieve_scores'] = scores_example
previous_result['retrieval_f1'] = f1_example
previous_result['retrieval_recall'] = recall_example


def base_passage_filter_test(contents, ids, scores):
    assert len(contents) == len(ids) == len(scores) == 2
    for content_list, id_list, score_list in zip(contents, ids, scores):
        assert isinstance(content_list, list)
        assert isinstance(id_list, list)
        assert isinstance(score_list, list)
        for content, _id, score in zip(content_list, id_list, score_list):
            assert isinstance(content, str)
            assert isinstance(_id, str)
            assert isinstance(score, float)

    assert all([len(c) > 0 for c in contents])
    assert all([len(i) > 0 for i in ids])
    assert all([len(s) > 0 for s in scores])


def base_passage_filter_node_test(result_df):
    assert all([column_name in result_df.columns for column_name in
                ["retrieved_contents", "retrieved_ids", "retrieve_scores"]])
    base_passage_filter_test(result_df['retrieved_contents'].tolist(), result_df['retrieved_ids'].tolist(),
                             result_df['retrieve_scores'].tolist())



import asyncio
import os
from typing import List, Tuple, Optional

import aiohttp

from autorag.nodes.passagereranker.base import passage_reranker_node
from autorag.utils.util import process_batch

JINA_API_URL = "https://api.jina.ai/v1/rerank"


@passage_reranker_node
def jina_reranker(queries: List[str], contents_list: List[List[str]],
                  scores_list: List[List[float]], ids_list: List[List[str]],
                  top_k: int, api_key: Optional[str] = None,
                  model: str = "jina-reranker-v1-base-en",
                  batch: int = 8
                  ) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank a list of contents with Jina rerank models.
    You can get the API key from https://jina.ai/reranker and set it in the environment variable JINAAI_API_KEY.

    :param queries: The list of queries to use for reranking
    :param contents_list: The list of lists of contents to rerank
    :param scores_list: The list of lists of scores retrieved from the initial ranking
    :param ids_list: The list of lists of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param api_key: The API key for Jina rerank.
        You can set it in the environment variable JINAAI_API_KEY.
        Or, you can directly set it on the config YAML file using this parameter.
        Default is env variable "JINAAI_API_KEY".
    :param model: The model name for Cohere rerank.
        You can choose between "jina-reranker-v1-base-en" and "jina-colbert-v1-en".
        Default is "jina-reranker-v1-base-en".
    :param batch: The number of queries to be processed in a batch
    :return: Tuple of lists containing the reranked contents, ids, and scores
    """
    if api_key is None:
        api_key = os.getenv("JINAAI_API_KEY", None)
        if api_key is None:
            raise ValueError("API key is not provided."
                             "You can set it as an argument or as an environment variable 'JINAAI_API_KEY'")

    tasks = [jina_reranker_pure(query, contents, scores, ids, top_k=top_k, api_key=api_key, model=model) for
             query, contents, scores, ids in
             zip(queries, contents_list, scores_list, ids_list)]
    loop = asyncio.get_event_loop()
    results = loop.run_until_complete(process_batch(tasks, batch))

    content_result, id_result, score_result = zip(*results)

    return list(content_result), list(id_result), list(score_result)


async def jina_reranker_pure(query: str, contents: List[str],
                             scores: List[float], ids: List[str],
                             top_k: int, api_key: str,
                             model: str = "jina-reranker-v1-base-en") -> Tuple[List[str], List[str], List[float]]:
    session = aiohttp.ClientSession()
    session.headers.update({"Authorization": f"Bearer {api_key}", "Accept-Encoding": "identity"})
    async with session.post(
            JINA_API_URL,
            json={
                "query": query,
                "documents": contents,
                "model": model,
                "top_n": top_k,
            },
    ) as resp:
        resp_json = await resp.json()
        if 'results' not in resp_json:
            raise RuntimeError(f"Invalid response from Jina API: {resp_json['detail']}")

        results = resp_json['results']
        indices = list(map(lambda x: x['index'], results))
        score_result = list(map(lambda x: x['relevance_score'], results))
        id_result = list(map(lambda x: ids[x], indices))
        content_result = list(map(lambda x: contents[x], indices))

        await session.close()
        return content_result, id_result, score_result



import importlib.resources
import logging
import os
import pathlib
import subprocess
from pathlib import Path
from typing import Optional

import click

from autorag.deploy import Runner
from autorag.deploy import extract_best_config as original_extract_best_config
from autorag.evaluator import Evaluator

logger = logging.getLogger("AutoRAG")


@click.group()
def cli():
    pass


@click.command()
@click.option('--config', '-c', help='Path to config yaml file. Must be yaml or yml file.', type=str)
@click.option('--qa_data_path', help='Path to QA dataset. Must be parquet file.', type=str)
@click.option('--corpus_data_path', help='Path to corpus dataset. Must be parquet file.', type=str)
@click.option('--project_dir', help='Path to project directory.', type=str, default=None)
def evaluate(config, qa_data_path, corpus_data_path, project_dir):
    if not config.endswith('.yaml') and not config.endswith('.yml'):
        raise ValueError(f"Config file {config} is not a parquet file.")
    if not os.path.exists(config):
        raise ValueError(f"Config file {config} does not exist.")
    evaluator = Evaluator(qa_data_path, corpus_data_path, project_dir=project_dir)
    evaluator.start_trial(config)
    logger.info('Evaluation complete.')


@click.command()
@click.option('--config_path', type=str, help='Path to extracted config yaml file.')
@click.option('--host', type=str, default='0.0.0.0', help='Host address')
@click.option('--port', type=int, default=8000, help='Port number')
@click.option('--project_dir', help='Path to project directory.', type=str, default=None)
def run_api(config_path, host, port, project_dir):
    runner = Runner.from_yaml(config_path, project_dir=project_dir)
    logger.info(f"Running API server at {host}:{port}...")
    runner.run_api_server(host, port)


@click.command()
@click.option('--yaml_path', type=click.Path(path_type=Path), help='Path to the YAML file.')
@click.option('--project_dir', type=click.Path(path_type=Path), help='Path to the project directory.')
@click.option('--trial_path', type=click.Path(path_type=Path), help='Path to the trial directory.')
def run_web(yaml_path: Optional[str], project_dir: Optional[str], trial_path: Optional[str]):
    try:
        with importlib.resources.path('autorag', 'web.py') as web_path:
            web_py_path = str(web_path)
    except ImportError:
        raise ImportError("Could not locate the web.py file within the autorag package."
                          " Please ensure that autorag is correctly installed.")

    if not yaml_path and not trial_path:
        raise ValueError("yaml_path or trial_path must be given.")
    elif yaml_path and trial_path:
        raise ValueError("yaml_path and trial_path cannot be given at the same time.")
    elif yaml_path and not project_dir:
        subprocess.run(['streamlit', 'run', web_py_path, '--', '--yaml_path', yaml_path])
    elif yaml_path and project_dir:
        subprocess.run(['streamlit', 'run', web_py_path, '--', '--yaml_path', yaml_path, '--project_dir', project_dir])
    elif trial_path:
        subprocess.run(['streamlit', 'run', web_py_path, '--', '--trial_path', trial_path])


@click.command()
@click.option('--trial_path', type=click.Path(), help='Path to the trial directory.')
@click.option('--output_path', type=click.Path(), help='Path to the output directory.'
                                                       ' Must be .yaml or .yml file.')
def extract_best_config(trial_path: str, output_path: str):
    original_extract_best_config(trial_path, output_path)


@click.command()
@click.option('--trial_path', help='Path to trial directory.', type=str)
def restart_evaluate(trial_path):
    if not os.path.exists(trial_path):
        raise ValueError(f"trial_path {trial_path} does not exist.")
    project_dir = pathlib.PurePath(trial_path).parent
    qa_data_path = os.path.join(project_dir, 'data', 'qa.parquet')
    corpus_data_path = os.path.join(project_dir, 'data', 'corpus.parquet')
    evaluator = Evaluator(qa_data_path, corpus_data_path, project_dir)
    evaluator.restart_trial(trial_path)
    logger.info('Evaluation complete.')


cli.add_command(evaluate, 'evaluate')
cli.add_command(run_api, 'run_api')
cli.add_command(run_web, 'run_web')
cli.add_command(extract_best_config, 'extract_best_config')
cli.add_command(restart_evaluate, 'restart_evaluate')

if __name__ == '__main__':
    cli()



import pytest

from autorag.nodes.retrieval.hybrid_cc import cc_pure, hybrid_cc
from tests.autorag.nodes.retrieval.test_hybrid_base import (sample_ids_2, sample_scores_2, sample_ids_3,
                                                            sample_scores_3, pseudo_project_dir,
                                                            base_hybrid_weights_node_test, sample_ids_non_overlap)


def test_cc_pure():
    result_id, result_scores = cc_pure(sample_ids_2, sample_scores_2,
                                       weights=(0.3, 0.7), top_k=3)
    assert result_scores == pytest.approx([1.0, 0.23792372, 0.175])
    assert result_id == ['id-1', 'id-4', 'id-2']


def test_cc_non_overlap():
    result_id, result_scores = cc_pure(sample_ids_non_overlap, sample_scores_2,
                                       weights=(0.3, 0.7), top_k=3)
    assert result_scores == pytest.approx([0.7, 0.3, 0.23792372])
    assert result_id == ['id-6', 'id-1', 'id-4']


def test_hybrid_cc():
    result_id, result_scores = hybrid_cc.__wrapped__(sample_ids_3, sample_scores_3, top_k=3, weights=(0.3, 0.7))
    assert result_id[0] == ['id-1', 'id-4', 'id-2']
    assert result_id[1] == ['id-2', 'id-5', 'id-3']
    assert result_scores[0] == pytest.approx([1.0, 0.23792372, 0.175])
    assert result_scores[1] == pytest.approx([1.0, 0.23792372, 0.175])


def test_hybrid_cc_node(pseudo_project_dir):
    retrieve_scores = [1.0, 0.23792372, 0.175]
    base_hybrid_weights_node_test(hybrid_cc, pseudo_project_dir, retrieve_scores)



import asyncio
from typing import List, Tuple

import chromadb
import pandas as pd
from llama_index.core.embeddings import BaseEmbedding
from llama_index.embeddings.openai import OpenAIEmbedding

from autorag.nodes.retrieval.base import retrieval_node, evenly_distribute_passages
from autorag.utils import validate_corpus_dataset
from autorag.utils.util import process_batch, openai_truncate_by_token


@retrieval_node
def vectordb(queries: List[List[str]], top_k: int, collection: chromadb.Collection,
             embedding_model: BaseEmbedding,
             batch: int = 128) -> Tuple[List[List[str]], List[List[float]]]:
    """
    VectorDB retrieval function.
    You have to get a chroma collection that is already ingested.
    You have to get an embedding model that is already used in ingesting.

    :param queries: 2-d list of query strings.
        Each element of the list is a query strings of each row.
    :param top_k: The number of passages to be retrieved.
    :param collection: A chroma collection instance that will be used to retrieve passages.
    :param embedding_model: An embedding model instance that will be used to embed queries.
    :param batch: The number of queries to be processed in parallel.
        This is used to prevent API error at the query embedding.
        Default is 128.

    :return: The 2-d list contains a list of passage ids that retrieved from vectordb and 2-d list of its scores.
        It will be a length of queries. And each element has a length of top_k.
    """
    # check if bm25_corpus is valid
    assert (collection.count() > 0), \
        "collection must contain at least one document. Please check you ingested collection correctly."
    # run async vector_db_pure function
    tasks = [vectordb_pure(input_queries, top_k, collection, embedding_model) for input_queries in queries]
    loop = asyncio.get_event_loop()
    results = loop.run_until_complete(process_batch(tasks, batch_size=batch))
    id_result = list(map(lambda x: x[0], results))
    score_result = list(map(lambda x: x[1], results))
    return id_result, score_result


async def vectordb_pure(queries: List[str], top_k: int, collection: chromadb.Collection,
                        embedding_model: BaseEmbedding) -> Tuple[List[str], List[float]]:
    """
    Async VectorDB retrieval function.
    Its usage is for async retrieval of vector_db row by row.

    :param queries: A list of query strings.
    :param top_k: The number of passages to be retrieved.
    :param collection: A chroma collection instance that will be used to retrieve passages.
    :param embedding_model: An embedding model instance that will be used to embed queries.

    :return: The tuple contains a list of passage ids that retrieved from vectordb and a list of its scores.
    """
    # embed query
    embedded_queries = list(map(embedding_model.get_query_embedding, queries))

    id_result, score_result = [], []
    for embedded_query in embedded_queries:
        result = collection.query(query_embeddings=embedded_query, n_results=top_k)
        id_result.extend(result['ids'])
        score_result.extend(result['distances'])

    # Distribute passages evenly
    id_result, score_result = evenly_distribute_passages(id_result, score_result, top_k)
    # sort id_result and score_result by score
    result = [(_id, score) for score, _id in
              sorted(zip(score_result, id_result), key=lambda pair: pair[0], reverse=True)]
    id_result, score_result = zip(*result)
    return list(id_result), list(score_result)


def vectordb_ingest(collection: chromadb.Collection, corpus_data: pd.DataFrame, embedding_model: BaseEmbedding,
                    batch: int = 128):
    embedding_model.embed_batch_size = batch
    validate_corpus_dataset(corpus_data)
    ids = corpus_data['doc_id'].tolist()

    # Query the collection to check if IDs already exist
    existing_ids = set(collection.get(ids=ids)['ids'])  # Assuming 'ids' is the key in the response
    new_passage = corpus_data[~corpus_data['doc_id'].isin(existing_ids)]

    if not new_passage.empty:
        new_contents = new_passage['contents'].tolist()

        # truncate by token if embedding_model is OpenAIEmbedding
        if isinstance(embedding_model, OpenAIEmbedding):
            openai_embedding_limit = 8191
            new_contents = openai_truncate_by_token(new_contents, openai_embedding_limit, embedding_model.model_name)

        new_ids = new_passage['doc_id'].tolist()
        embedded_contents = embedding_model.get_text_embedding_batch(new_contents, show_progress=True)
        collection.add(ids=new_ids, embeddings=embedded_contents)



import asyncio
from typing import List, Tuple

import torch
from FlagEmbedding import FlagReranker

from autorag.nodes.passagereranker.base import passage_reranker_node
from autorag.utils.util import process_batch


@passage_reranker_node
def flag_embedding_reranker(queries: List[str], contents_list: List[List[str]],
                            scores_list: List[List[float]], ids_list: List[List[str]],
                            top_k: int, batch: int = 64, use_fp16: bool = False,
                            model_name: str = "BAAI/bge-reranker-large",
                            ) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank a list of contents based on their relevance to a query using BAAI normal-Reranker model.

    :param queries: The list of queries to use for reranking
    :param contents_list: The list of lists of contents to rerank
    :param scores_list: The list of lists of scores retrieved from the initial ranking
    :param ids_list: The list of lists of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param batch: The number of queries to be processed in a batch
    :param use_fp16: Whether to use fp16 for inference
    :param model_name: The name of the BAAI Reranker normal-model name.
        Default is "BAAI/bge-reranker-large"
    :return: tuple of lists containing the reranked contents, ids, and scores
    """
    model = FlagReranker(
        model_name_or_path=model_name, use_fp16=use_fp16
    )
    tasks = [flag_embedding_reranker_pure(query, contents, scores, top_k, ids, model)
             for query, contents, scores, ids in zip(queries, contents_list, scores_list, ids_list)]
    loop = asyncio.get_event_loop()
    results = loop.run_until_complete(process_batch(tasks, batch_size=batch))
    content_result = list(map(lambda x: x[0], results))
    id_result = list(map(lambda x: x[1], results))
    score_result = list(map(lambda x: x[2], results))

    del model
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    return content_result, id_result, score_result


async def flag_embedding_reranker_pure(query: str, contents: List[str], scores: List[float], top_k: int,
                                       ids: List[str], model) -> Tuple[List[str], List[str], List[float]]:
    """
    Rerank a list of contents based on their relevance to a query using BAAI Reranker model.

    :param query: The query to use for reranking
    :param contents: The list of contents to rerank
    :param scores: The list of scores retrieved from the initial ranking
    :param ids: The list of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param model: BAAI Reranker model.
    :return: tuple of lists containing the reranked contents, ids, and scores
    """
    input_texts = [(query, content) for content in contents]
    with torch.no_grad():
        pred_scores = model.compute_score(sentence_pairs=input_texts)

    content_ids_probs = list(zip(contents, ids, pred_scores))

    # Sort the list of pairs based on the relevance score in descending order
    sorted_content_ids_probs = sorted(content_ids_probs, key=lambda x: x[2], reverse=True)

    # crop with top_k
    if len(contents) < top_k:
        top_k = len(contents)
    sorted_content_ids_probs = sorted_content_ids_probs[:top_k]

    content_result, id_result, score_result = zip(*sorted_content_ids_probs)

    return list(content_result), list(id_result), list(score_result)



from .generation import bleu, meteor, rouge, sem_score, g_eval, bert_score
from .retrieval import retrieval_f1, retrieval_recall, retrieval_precision
from .retrieval_contents import retrieval_token_f1, retrieval_token_precision, retrieval_token_recall



import os
import pathlib
from typing import Dict, List, Optional

import pandas as pd

from autorag.schema import Node
from autorag.utils.util import load_summary_file


def make_node_lines(node_line_dict: Dict) -> List[Node]:
    """
    This method makes a list of nodes from node line dictionary.
    :param node_line_dict: Node_line_dict loaded from yaml file, or get from user input.
    :return: List of Nodes inside this node line.
    """
    nodes = node_line_dict.get('nodes')
    if nodes is None:
        raise ValueError("Node line must have \'nodes\' key.")
    node_objects = list(map(lambda x: Node.from_dict(x), nodes))
    return node_objects


def run_node_line(nodes: List[Node],
                  node_line_dir: str,
                  previous_result: Optional[pd.DataFrame] = None):
    """
    Run the whole node line by running each node.

    :param nodes: A list of nodes.
    :param node_line_dir: This node line's directory.
    :param previous_result: A result of the previous node line.
        If None, it loads qa data from data/qa.parquet.
    :return: The final result of the node line.
    """
    if previous_result is None:
        project_dir = pathlib.PurePath(node_line_dir).parent.parent
        qa_path = os.path.join(project_dir, "data", "qa.parquet")
        if not os.path.exists(qa_path):
            raise ValueError(f"qa.parquet does not exist in {qa_path}.")
        previous_result = pd.read_parquet(qa_path)

    summary_lst = []
    for node in nodes:
        previous_result = node.run(previous_result, node_line_dir)
        node_summary_df = load_summary_file(os.path.join(node_line_dir, node.node_type, 'summary.csv'))
        best_node_row = node_summary_df.loc[node_summary_df['is_best']]
        summary_lst.append({
            'node_type': node.node_type,
            'best_module_filename': best_node_row['filename'].values[0],
            'best_module_name': best_node_row['module_name'].values[0],
            'best_module_params': best_node_row['module_params'].values[0],
            'best_execution_time': best_node_row['execution_time'].values[0],
        })
    pd.DataFrame(summary_lst).to_csv(os.path.join(node_line_dir, 'summary.csv'), index=False)
    return previous_result



import asyncio
from typing import List, Tuple

import torch
from FlagEmbedding import FlagLLMReranker

from autorag.nodes.passagereranker.base import passage_reranker_node
from autorag.nodes.passagereranker.flag_embedding import flag_embedding_reranker_pure
from autorag.utils.util import process_batch


@passage_reranker_node
def flag_embedding_llm_reranker(queries: List[str], contents_list: List[List[str]],
                                scores_list: List[List[float]], ids_list: List[List[str]],
                                top_k: int, batch: int = 64, use_fp16: bool = False,
                                model_name: str = "BAAI/bge-reranker-v2-gemma",
                                ) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank a list of contents based on their relevance to a query using BAAI LLM-based-Reranker model.

    :param queries: The list of queries to use for reranking
    :param contents_list: The list of lists of contents to rerank
    :param scores_list: The list of lists of scores retrieved from the initial ranking
    :param ids_list: The list of lists of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param batch: The number of queries to be processed in a batch
    :param use_fp16: Whether to use fp16 for inference
    :param model_name: The name of the BAAI Reranker LLM-based-model name.
        Default is "BAAI/bge-reranker-v2-gemma"
    :return: tuple of lists containing the reranked contents, ids, and scores
    """
    model = FlagLLMReranker(
        model_name_or_path=model_name, use_fp16=use_fp16
    )
    tasks = [flag_embedding_reranker_pure(query, contents, scores, top_k, ids, model)
             for query, contents, scores, ids in zip(queries, contents_list, scores_list, ids_list)]
    loop = asyncio.get_event_loop()
    results = loop.run_until_complete(process_batch(tasks, batch_size=batch))
    content_result = list(map(lambda x: x[0], results))
    id_result = list(map(lambda x: x[1], results))
    score_result = list(map(lambda x: x[2], results))

    del model
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    return content_result, id_result, score_result



import mimetypes
import os
from datetime import datetime
from pathlib import Path
from typing import Dict, List

import pandas as pd
from langchain_core.documents import Document


def get_file_metadata(file_path: str) -> Dict:
    """Get some handy metadate from filesystem.

    Args:
        file_path: str: file path in str
    """
    return {
        "file_path": file_path,
        "file_name": os.path.basename(file_path),
        "file_type": mimetypes.guess_type(file_path)[0],
        "file_size": os.path.getsize(file_path),
        "creation_datetime": datetime.fromtimestamp(
            Path(file_path).stat().st_ctime
        ).strftime("%Y-%m-%d"),
        "last_modified_datetime": datetime.fromtimestamp(
            Path(file_path).stat().st_mtime
        ).strftime("%Y-%m-%d"),
        "last_accessed_datetime": datetime.fromtimestamp(
            Path(file_path).stat().st_atime
        ).strftime("%Y-%m-%d"),
    }


def add_essential_metadata(metadata: Dict) -> Dict:
    if 'last_modified_datetime' not in metadata:
        metadata['last_modified_datetime'] = datetime.now()
    return metadata


def corpus_df_to_langchain_documents(corpus_df: pd.DataFrame) -> List[Document]:
    page_contents = corpus_df['contents'].tolist()
    ids = corpus_df['doc_id'].tolist()
    metadatas = corpus_df['metadata'].tolist()
    return list(map(lambda x: Document(page_content=x[0], metadata={'filename': x[1], **x[2]}),
                    zip(page_contents, ids, metadatas)))



from autorag import generator_models
from autorag.nodes.queryexpansion import hyde
from tests.autorag.nodes.queryexpansion.test_query_expansion_base import project_dir, previous_result, \
    base_query_expansion_node_test, ingested_vectordb_node
from tests.mock import MockLLM

sample_query = ["How many members are in Newjeans?", "What is visconde structure?"]


def test_hyde():
    llm = MockLLM()
    original_hyde = hyde.__wrapped__
    result = original_hyde(sample_query, llm, prompt="")
    assert len(result[0]) == 1
    assert len(result) == 2


def test_hyde_node(ingested_vectordb_node):
    generator_models['mock'] = MockLLM
    result_df = hyde(project_dir=project_dir, previous_result=previous_result,
                     llm="mock", max_tokens=64)
    base_query_expansion_node_test(result_df)



import os
import pathlib
import tempfile

import pandas as pd
import pytest
from guidance import models
from llama_index.core import SimpleDirectoryReader

from autorag.data.corpus.llama_index import llama_documents_to_parquet
from autorag.data.qacreation.simple import generate_simple_qa_dataset, generate_qa_row
from autorag.data.utils.util import get_file_metadata
from autorag.utils.preprocess import validate_qa_dataset, validate_corpus_dataset

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent.parent.parent

raw_dir = os.path.join(root_dir, "resources", "data_creation", "raw_dir")

load_file_name = "test_corpus.parquet"


@pytest.fixture
def load_dir():
    with tempfile.TemporaryDirectory() as temp_dir:
        yield temp_dir


@pytest.fixture
def output_filedir():
    with tempfile.TemporaryDirectory() as temp_dir:
        yield temp_dir


def test_generate_simple_qa_dataset(load_dir, output_filedir):
    loader = SimpleDirectoryReader(
        file_metadata=get_file_metadata,
        input_dir=raw_dir
    )

    documents = loader.load_data(
        show_progress=True,
        num_workers=0
    )
    llama_documents_to_parquet(llama_documents=documents, output_filepath=os.path.join(load_dir, load_file_name))
    assert os.path.exists(os.path.join(load_dir, load_file_name))
    corpus_data = pd.read_parquet(os.path.join(load_dir, load_file_name))
    validate_corpus_dataset(corpus_data)

    qa_dataset = generate_simple_qa_dataset(corpus_data=pd.read_parquet(os.path.join(load_dir, load_file_name)),
                                            llm=models.OpenAI("gpt-3.5-turbo"),
                                            output_filepath=os.path.join(output_filedir, 'qa.parquet'),
                                            generate_row_function=generate_qa_row)
    validate_qa_dataset(qa_dataset)
    assert os.path.exists(os.path.join(output_filedir, 'qa.parquet'))



import asyncio
import os
import pickle
from typing import List, Dict, Tuple

import numpy as np
import pandas as pd
from rank_bm25 import BM25Okapi
from transformers import AutoTokenizer

from autorag.nodes.retrieval.base import retrieval_node, evenly_distribute_passages
from autorag.utils import validate_corpus_dataset


@retrieval_node
def bm25(queries: List[List[str]], top_k: int, bm25_corpus: Dict) -> Tuple[List[List[str
]], List[List[float]]]:
    """
    BM25 retrieval function.
    You have to load a pickle file that is already ingested.

    :param queries: 2-d list of query strings.
        Each element of the list is a query strings of each row.
    :param top_k: The number of passages to be retrieved.
    :param bm25_corpus: A dictionary containing the bm25 corpus, which is doc_id from corpus and tokenized corpus.
        Its data structure looks like this:

        .. Code:: python

            {
                "tokens": [], # 2d list of tokens
                "passage_id": [], # 2d list of passage_id.
            }

    :return: The 2-d list contains a list of passage ids that retrieved from bm25 and 2-d list of its scores.
        It will be a length of queries. And each element has a length of top_k.
    """
    # check if bm25_corpus is valid
    assert ("tokens" and "passage_id" in list(bm25_corpus.keys())), \
        "bm25_corpus must contain tokens and passage_id. Please check you ingested bm25 corpus correctly."
    tokenizer = AutoTokenizer.from_pretrained("gpt2", use_fast=False)
    bm25_instance = BM25Okapi(bm25_corpus["tokens"])
    # run async bm25_pure function
    tasks = [bm25_pure(input_queries, top_k, tokenizer, bm25_instance, bm25_corpus) for input_queries in queries]
    loop = asyncio.get_event_loop()
    results = loop.run_until_complete(asyncio.gather(*tasks))
    id_result = list(map(lambda x: x[0], results))
    score_result = list(map(lambda x: x[1], results))
    return id_result, score_result


async def bm25_pure(queries: List[str], top_k: int, tokenizer, bm25_api: BM25Okapi, bm25_corpus: Dict) -> Tuple[
        List[str], List[float]]:
    """
    Async BM25 retrieval function.
    Its usage is for async retrieval of bm25 row by row.

    :param queries: A list of query strings.
    :param top_k: The number of passages to be retrieved.
    :param tokenizer: A tokenizer that will be used to tokenize queries.
    :param bm25_api: A bm25 api instance that will be used to retrieve passages.
    :param bm25_corpus: A dictionary containing the bm25 corpus, which is doc_id from corpus and tokenized corpus.
        Its data structure looks like this:

        .. Code:: python

            {
                "tokens": [], # 2d list of tokens
                "passage_id": [], # 2d list of passage_id. Type must be str.
            }
    :return: The tuple contains a list of passage ids that retrieved from bm25 and its scores.
    """
    # I don't make queries operation to async, because queries length might be small, so it will occur overhead.
    tokenized_queries = tokenizer(queries).input_ids
    id_result = []
    score_result = []
    for query in tokenized_queries:
        scores = bm25_api.get_scores(query)
        sorted_scores = sorted(scores, reverse=True)
        top_n_index = np.argsort(scores)[::-1][:top_k]
        ids = [bm25_corpus['passage_id'][i] for i in top_n_index]
        id_result.append(ids)
        score_result.append(sorted_scores[:top_k])

    # make a total result to top_k
    id_result, score_result = evenly_distribute_passages(id_result, score_result, top_k)
    # sort id_result and score_result by score
    result = [(_id, score) for score, _id in
              sorted(zip(score_result, id_result), key=lambda pair: pair[0], reverse=True)]
    id_result, score_result = zip(*result)
    return list(id_result), list(score_result)


def bm25_ingest(corpus_path: str, corpus_data: pd.DataFrame):
    if not corpus_path.endswith('.pkl'):
        raise ValueError(f"Corpus path {corpus_path} is not a pickle file.")
    validate_corpus_dataset(corpus_data)
    ids = corpus_data['doc_id'].tolist()

    # Initialize bm25_corpus
    bm25_corpus = pd.DataFrame()

    # Load the BM25 corpus if it exists and get the passage ids
    if os.path.exists(corpus_path) and os.path.getsize(corpus_path) > 0:
        with open(corpus_path, 'rb') as r:
            corpus = pickle.load(r)
            bm25_corpus = pd.DataFrame.from_dict(corpus)
        duplicated_passage_rows = bm25_corpus[bm25_corpus['passage_id'].isin(ids)]
        new_passage = corpus_data[~corpus_data['doc_id'].isin(duplicated_passage_rows['passage_id'])]
    else:
        new_passage = corpus_data

    if not new_passage.empty:
        tokenizer = AutoTokenizer.from_pretrained('gpt2', use_fast=False)
        results = new_passage.swifter.apply(lambda x: bm25_tokenize(x['contents'], x['doc_id'], tokenizer), axis=1)
        tokenized_corpus, passage_ids = zip(*results)

        new_bm25_corpus = pd.DataFrame({
            'tokens': list(tokenized_corpus),
            'passage_id': list(passage_ids),
        })

        if not bm25_corpus.empty:
            bm25_corpus_updated = pd.concat([bm25_corpus, new_bm25_corpus], ignore_index=True)
            bm25_dict = bm25_corpus_updated.to_dict('list')
        else:
            bm25_dict = new_bm25_corpus.to_dict('list')

        with open(corpus_path, 'wb') as w:
            pickle.dump(bm25_dict, w)


def bm25_tokenize(queries: List[str], passage_id: str, tokenizer) -> Tuple[List[int], str]:
    tokenized_queries = tokenizer(queries).input_ids
    return tokenized_queries, passage_id



import asyncio
from typing import List

from llama_index.core.service_context_elements.llm_predictor import LLMPredictorType

from autorag.nodes.queryexpansion.base import query_expansion_node
from autorag.utils.util import process_batch

hyde_prompt = "Please write a passage to answer the question"


@query_expansion_node
def hyde(queries: List[str], llm: LLMPredictorType,
         prompt: str = hyde_prompt,
         batch: int = 16) -> List[List[str]]:
    """
    HyDE, which inspired by "Precise Zero-shot Dense Retrieval without Relevance Labels" (https://arxiv.org/pdf/2212.10496.pdf)
    LLM model creates a hypothetical passage.
    And then, retrieve passages using hypothetical passage as a query.
    :param queries: List[str], queries to retrieve.
    :param llm: llm to use for hypothetical passage generation.
    :param prompt: prompt to use when generating hypothetical passage
    :param batch: Batch size for llm.
        Default is 16.
    :return: List[List[str]], List of hyde results.
    """
    # Run async query_decompose_pure function
    tasks = [hyde_pure(query, llm, prompt) for query in queries]
    loop = asyncio.get_event_loop()
    results = loop.run_until_complete(process_batch(tasks, batch_size=batch))
    return results


async def hyde_pure(query: str, llm: LLMPredictorType,
                    prompt: str = hyde_prompt) -> List[str]:
    if prompt is "":
        prompt = hyde_prompt
    full_prompt = prompt + f"\nQuestion: {query}\nPassage:"
    hyde_answer = await llm.acomplete(full_prompt)
    return [hyde_answer.text]



import asyncio
from typing import List, Tuple

import torch
from transformers import AutoModel, AutoTokenizer

from autorag.nodes.passagereranker.base import passage_reranker_node
from autorag.utils.util import process_batch


@passage_reranker_node
def colbert_reranker(queries: List[str], contents_list: List[List[str]],
                     scores_list: List[List[float]], ids_list: List[List[str]],
                     top_k: int, batch: int = 64,
                     model_name: str = "colbert-ir/colbertv2.0",
                     ) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank a list of contents with Colbert rerank models.
    You can get more information about a Colbert model at https://huggingface.co/colbert-ir/colbertv2.0.
    It uses BERT-based model, so recommend using CUDA gpu for faster reranking.

    :param queries: The list of queries to use for reranking
    :param contents_list: The list of lists of contents to rerank
    :param scores_list: The list of lists of scores retrieved from the initial ranking
    :param ids_list: The list of lists of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param batch: The number of queries to be processed in a batch
        Default is 64.
    :param model_name: The model name for Colbert rerank.
        You can choose colbert model for reranking.
        Default is "colbert-ir/colbertv2.0".
    :return: Tuple of lists containing the reranked contents, ids, and scores
    """
    device = "cuda" if torch.cuda.is_available() else "cpu"
    model = AutoModel.from_pretrained(model_name).to(device)
    tokenizer = AutoTokenizer.from_pretrained(model_name)

    # Run async cohere_rerank_pure function
    tasks = [get_colbert_score(query, document, model, tokenizer) for query, document, ids in
             zip(queries, contents_list, ids_list)]
    loop = asyncio.get_event_loop()
    score_results = loop.run_until_complete(process_batch(tasks, batch_size=batch))

    del model

    def rerank_results(contents, ids, scores, top_k):
        reranked_content, reranked_id, reranked_score = zip(
            *sorted(zip(contents, ids, scores), key=lambda x: x[2], reverse=True))
        return list(reranked_content)[:top_k], list(reranked_id)[:top_k], list(reranked_score)[:top_k]

    reranked_contents_list, reranked_ids_list, reranked_scores_list = zip(*list(map(
        rerank_results, contents_list, ids_list, score_results, [top_k] * len(contents_list))))
    return list(reranked_contents_list), list(reranked_ids_list), list(reranked_scores_list)


async def get_colbert_score(query: str, content_list: List[str],
                            model, tokenizer) -> List[float]:
    query_encoding = tokenizer(query, return_tensors="pt")
    query_embedding = model(**query_encoding).last_hidden_state
    rerank_score_list = []

    for document_text in content_list:
        document_encoding = tokenizer(
            document_text, return_tensors="pt", truncation=True, max_length=512
        )
        document_embedding = model(**document_encoding).last_hidden_state

        sim_matrix = torch.nn.functional.cosine_similarity(
            query_embedding.unsqueeze(2), document_embedding.unsqueeze(1), dim=-1
        )

        # Take the maximum similarity for each query token (across all document tokens)
        # sim_matrix shape: [batch_size, query_length, doc_length]
        max_sim_scores, _ = torch.max(sim_matrix, dim=2)
        rerank_score_list.append(torch.mean(max_sim_scores, dim=1))

    return list(map(float, rerank_score_list))



import pandas as pd

from autorag import generator_models
from autorag.nodes.generator import llama_index_llm
from tests.autorag.nodes.generator.test_generator_base import (prompts, check_generated_texts, check_generated_tokens,
                                                               check_generated_log_probs)
from tests.mock import MockLLM


def test_llama_index_llm():
    llama_index_llm_original = llama_index_llm.__wrapped__
    answers, tokens, log_probs = llama_index_llm_original(prompts, MockLLM())
    check_generated_texts(answers)
    check_generated_tokens(tokens)
    check_generated_log_probs(log_probs)
    assert all(all(log_prob == 0.5 for log_prob in log_prob_list) for log_prob_list in log_probs)
    assert all(len(tokens[i]) == len(log_probs[i]) for i in range(len(tokens)))


def test_llama_index_llm_node():
    generator_models['mock'] = MockLLM
    previous_result = pd.DataFrame(
        {
            'prompts': prompts,
            'qid': ['id-1', 'id-2', 'id-3']
        })
    result_df = llama_index_llm(project_dir='.', previous_result=previous_result, llm='mock',
                                temperature=0.5, top_p=0.9)
    check_generated_texts(result_df['generated_texts'].tolist())
    check_generated_tokens(result_df['generated_tokens'].tolist())
    check_generated_log_probs(result_df['generated_log_probs'].tolist())



import functools
from typing import List


def retrieval_metric(func):
    @functools.wraps(func)
    def wrapper(retrieval_gt: List[List[List[str]]], pred_ids: List[List[str]]) -> List[float]:
        return list(map(lambda x: func(x[0], x[1]), zip(retrieval_gt, pred_ids)))

    return wrapper


@retrieval_metric
def retrieval_f1(gt: List[List[str]], pred: List[str]):
    """
    Compute f1 score for retrieval.

    :param gt: 2-d list of ground truth ids.
        It contains and/or connections between ids.
    :param pred: Prediction ids.
    :return: The f1 score.
    """
    recall_score = retrieval_recall.__wrapped__(gt, pred)
    precision_score = retrieval_precision.__wrapped__(gt, pred)
    if recall_score + precision_score == 0:
        return 0
    else:
        return 2 * (recall_score * precision_score) / (recall_score + precision_score)


@retrieval_metric
def retrieval_recall(gt: List[List[str]], pred: List[str]):
    gt_sets = [frozenset(g) for g in gt]
    pred_set = set(pred)
    hits = sum(any(pred_id in gt_set for pred_id in pred_set) for gt_set in gt_sets)
    recall = hits / len(gt) if len(gt) > 0 else 0.0
    return recall


@retrieval_metric
def retrieval_precision(gt: List[List[str]], pred: List[str]):
    gt_sets = [frozenset(g) for g in gt]
    pred_set = set(pred)
    hits = sum(any(pred_id in gt_set for gt_set in gt_sets) for pred_id in pred_set)
    precision = hits / len(pred) if len(pred) > 0 else 0.0
    return precision



import asyncio

from llama_index.core import QueryBundle
from llama_index.core.schema import NodeWithScore, TextNode
from llama_index.llms.openai import OpenAI

from autorag.nodes.passagereranker import rankgpt
from autorag.nodes.passagereranker.rankgpt import AsyncRankGPTRerank
from tests.autorag.nodes.passagereranker.test_passage_reranker_base import queries_example, contents_example, \
    scores_example, ids_example, base_reranker_test, project_dir, previous_result, base_reranker_node_test


def test_async_rankgpt_rerank():
    query = queries_example[0]
    query_bundle = QueryBundle(query_str=query)
    nodes = list(map(lambda x: NodeWithScore(node=TextNode(text=x)), contents_example[0]))

    reranker = AsyncRankGPTRerank(top_n=3, llm=OpenAI())
    result, id_result = asyncio.run(reranker.async_postprocess_nodes(nodes, query_bundle))

    assert len(result) == 3
    assert all(isinstance(node, NodeWithScore) for node in result)


def test_rankgpt_reranker():
    top_k = 3
    original_rankgpt_reranker = rankgpt.__wrapped__
    contents_result, id_result, score_result \
        = original_rankgpt_reranker(queries_example, contents_example, scores_example, ids_example, top_k,
                                    llm=OpenAI(model="gpt-3.5-turbo-16k"))
    base_reranker_test(contents_result, id_result, score_result, top_k)


def test_rankgpt_node():
    top_k = 1
    result_df = rankgpt(project_dir=project_dir, previous_result=previous_result, top_k=top_k,
                        llm='openai', model='gpt-3.5-turbo', temperature=0.5, batch=8, verbose=True)
    base_reranker_node_test(result_df, top_k)

    top_k = 2
    result_df = rankgpt(project_dir=project_dir, previous_result=previous_result, top_k=top_k, batch=4)
    base_reranker_node_test(result_df, top_k)



from autorag import generator_models
from autorag.nodes.queryexpansion import query_decompose
from tests.autorag.nodes.queryexpansion.test_query_expansion_base import project_dir, previous_result, \
    base_query_expansion_node_test, ingested_vectordb_node
from tests.mock import MockLLM

sample_query = ["Which group has more members, Newjeans or Aespa?", "Which group has more members, STAYC or Aespa?"]


def test_query_decompose():
    llm = MockLLM(temperature=0.2)
    original_query_decompose = query_decompose.__wrapped__
    result = original_query_decompose(sample_query, llm, prompt="")
    assert len(result[0]) > 1
    assert len(result) == 2
    assert isinstance(result[0][0], str)


def test_query_decompose_node(ingested_vectordb_node):
    generator_models['mock'] = MockLLM
    result_df = query_decompose(project_dir=project_dir, previous_result=previous_result,
                                llm="mock", temperature=0.2)
    base_query_expansion_node_test(result_df)



import os
import pathlib
import tempfile

import pandas as pd
import pytest
from llama_index.llms.openai import OpenAI

from autorag.data.qacreation import make_single_content_qa, generate_qa_llama_index
from autorag.utils import validate_qa_dataset

root_dir = pathlib.PurePath(os.path.dirname(os.path.realpath(__file__))).parent.parent.parent
resource_dir = os.path.join(root_dir, "resources")


@pytest.fixture
def qa_parquet_filepath():
    with tempfile.NamedTemporaryFile(suffix=".parquet") as f:
        yield f.name


def test_single_content_qa(qa_parquet_filepath):
    corpus_df = pd.read_parquet(os.path.join(resource_dir, "corpus_data_sample.parquet"))
    qa_df = make_single_content_qa(
        corpus_df,
        content_size=3,
        qa_creation_func=generate_qa_llama_index,
        output_filepath=qa_parquet_filepath,
        llm=OpenAI(model="gpt-3.5-turbo", temperature=1.0),
        question_num_per_content=2,
        upsert=True,
    )
    validate_qa_dataset(qa_df)
    assert len(qa_df) == 6
    assert qa_df['retrieval_gt'].tolist()[0] == qa_df['retrieval_gt'].tolist()[1]
    assert qa_df['query'].tolist()[0] != qa_df['query'].tolist()[1]
    assert qa_df['generation_gt'].tolist()[0] != qa_df['generation_gt'].tolist()[1]

    assert all([len(x) == 1 and len(x[0]) == 1 for x in qa_df['retrieval_gt'].tolist()])
    assert all([len(x) == 1 for x in qa_df['generation_gt'].tolist()])



import asyncio
from typing import List, Tuple

import torch
from sentence_transformers import CrossEncoder

from autorag.nodes.passagereranker.base import passage_reranker_node
from autorag.utils.util import process_batch


@passage_reranker_node
def sentence_transformer_reranker(queries: List[str], contents_list: List[List[str]],
                                  scores_list: List[List[float]], ids_list: List[List[str]],
                                  top_k: int, batch: int = 64, sentence_transformer_max_length: int = 512,
                                  model_name: str = "cross-encoder/ms-marco-MiniLM-L-2-v2",
                                  ) -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank a list of contents based on their relevance to a query using Sentence Transformer model.

    :param queries: The list of queries to use for reranking
    :param contents_list: The list of lists of contents to rerank
    :param scores_list: The list of lists of scores retrieved from the initial ranking
    :param ids_list: The list of lists of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param batch: The number of queries to be processed in a batch
    :param sentence_transformer_max_length: The maximum length of the input text for the Sentence Transformer model
    :param model_name: The name of the Sentence Transformer model to use for reranking
        Default is "cross-encoder/ms-marco-MiniLM-L-2-v2"
    :return: tuple of lists containing the reranked contents, ids, and scores
    """
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model = CrossEncoder(
        model_name, max_length=sentence_transformer_max_length, device=device
    )
    tasks = [sentence_transformer_reranker_pure(query, contents, scores, top_k, ids, model)
             for query, contents, scores, ids in zip(queries, contents_list, scores_list, ids_list)]
    loop = asyncio.get_event_loop()
    results = loop.run_until_complete(process_batch(tasks, batch_size=batch))
    content_result = list(map(lambda x: x[0], results))
    id_result = list(map(lambda x: x[1], results))
    score_result = list(map(lambda x: x[2], results))

    del model
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    return content_result, id_result, score_result


async def sentence_transformer_reranker_pure(query: str, contents: List[str], scores: List[float], top_k: int,
                                             ids: List[str], model) -> Tuple[List[str], List[str], List[float]]:
    """
    Rerank a list of contents based on their relevance to a query using Sentence Transformer model.

    :param query: The query to use for reranking
    :param contents: The list of contents to rerank
    :param scores: The list of scores retrieved from the initial ranking
    :param ids: The list of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param model: The name of the Sentence Transformer model to use for reranking
    :return: tuple of lists containing the reranked contents, ids, and scores
    """
    input_texts = [(query, content) for content in contents]
    with torch.no_grad():
        pred_scores = model.predict(sentences=input_texts, apply_softmax=True)

    content_ids_probs = list(zip(contents, ids, pred_scores.tolist()))

    # Sort the list of pairs based on the relevance score in descending order
    sorted_content_ids_probs = sorted(content_ids_probs, key=lambda x: x[2], reverse=True)

    # crop with top_k
    if len(contents) < top_k:
        top_k = len(contents)
    sorted_content_ids_probs = sorted_content_ids_probs[:top_k]

    content_result, id_result, score_result = zip(*sorted_content_ids_probs)

    return list(content_result), list(id_result), list(score_result)



import asyncio
import os.path
import random
from typing import Optional, List, Dict, Any

from llama_index.core.service_context_elements.llm_predictor import LLMPredictorType

from autorag.utils.util import process_batch

package_dir = os.path.dirname(os.path.realpath(__file__))


def generate_qa_llama_index(
        llm: LLMPredictorType,
        contents: List[str],
        prompt: Optional[str] = None,
        question_num_per_content: int = 1,
        max_retries: int = 3,
        batch: int = 4,
) -> List[List[Dict]]:
    """
    Generate a qa set from the list of contents.
    It uses a single prompt for all contents.
    If you want to use more than one prompt for generating qa,
    you can consider using generate_qa_llama_index_by_ratio.

    :param llm: Llama index model
    :param contents: List of content strings.
    :param prompt: The prompt to use for the qa generation.
        The prompt must include the following placeholders:
        - {{text}}: The content string
        - {{num_questions}}: The number of questions to generate
        As default, the prompt is set to the default prompt for the question type.
    :param question_num_per_content: Number of questions to generate for each content.
        Default is 1.
    :param max_retries: The maximum number of retries when generated question number is not equal to the target number.
        Default is 3.
    :param batch: The batch size to process asynchronously.
        Default is 4.
    :return: 2-d list of dictionaries containing the query and generation_gt.
    """
    # load default prompt
    if prompt is None:
        prompt = open(os.path.join(package_dir, "llama_index_default_prompt.txt"), 'r').read()

    tasks = [
        async_qa_gen_llama_index(content, llm, prompt, question_num_per_content, max_retries)
        for content in contents
    ]
    loops = asyncio.get_event_loop()
    results = loops.run_until_complete(process_batch(tasks, batch))
    return results


def generate_qa_llama_index_by_ratio(
        llm: LLMPredictorType,
        contents: List[str],
        prompts_ratio: Dict,
        question_num_per_content: int = 1,
        max_retries: int = 3,
        random_state: int = 42,
        batch: int = 4,
) -> List[List[Dict]]:
    """
    Generate a qa set from the list of contents.
    You can set the ratio of prompts that you want to use for generating qa.
    It distributes the number of questions to generate for each content by the ratio randomly.

    :param llm: Llama index model
    :param contents: List of content strings.
    :param prompts_ratio: Dictionary of prompt paths and their ratios.
        Example: {"prompt/prompt1.txt": 0.5, "prompt/prompt2.txt": 0.5}
        The value sum doesn't have to be 1.
        The path must be the absolute path, and the file must exist.
        Plus, it has to be a text file which contains proper prompt.
        Each prompt must contain the following placeholders:
        - {{text}}: The content string
        - {{num_questions}}: The number of questions to generate
    :param question_num_per_content: Number of questions to generate for each content.
        Default is 1.
    :param max_retries: The maximum number of retries when generated question number is not equal to the target number.
        Default is 3.
    :param random_state: Random seed
        Default is 42.
    :param batch: The batch size to process asynchronously.
        Default is 4.
    :return: 2-d list of dictionaries containing the query and generation_gt.
    """
    prompts = list(map(lambda path: open(path, 'r').read(), prompts_ratio.keys()))
    assert all([validate_llama_index_prompt(prompt) for prompt in prompts])

    random.seed(random_state)
    random.shuffle(contents)

    slice_contents: List[List[str]] = distribute_list_by_ratio(contents, list(prompts_ratio.values()))
    tasks = [
        async_qa_gen_llama_index(content, llm, prompt, question_num_per_content, max_retries)
        for prompt, type_contents in zip(prompts, slice_contents)
        for content in type_contents
    ]

    loops = asyncio.get_event_loop()
    results = loops.run_until_complete(process_batch(tasks, batch))

    return results


async def async_qa_gen_llama_index(
        content: str,
        llm: LLMPredictorType,
        prompt: str,
        question_num: int = 1,
        max_retries: int = 3
):
    """
    Generate a qa set by using the given content and the llama index model.
    You must select the question type.

    :param content: Content string
    :param llm: Llama index model
    :param prompt: The prompt to use for the qa generation.
        The prompt must include the following placeholders:
        - {{text}}: The content string
        - {{num_questions}}: The number of questions to generate
    :param question_num: The number of questions to generate
    :param max_retries: Maximum number of retries when generated question number is not equal to the target number
    :return: List of dictionaries containing the query and generation_gt
    """
    validate_llama_index_prompt(prompt)

    async def generate(content: str, llm: LLMPredictorType):
        for _ in range(max_retries):
            output = await llm.acomplete(
                prompt.replace("{{text}}", content).replace("{{num_questions}}", str(question_num)))
            result = parse_output(output.text)
            if len(result) == question_num:
                return result
        raise InterruptedError(f"Failed to generate output of length {question_num} after {max_retries} retries.")

    return await generate(content, llm)


def validate_llama_index_prompt(prompt: str) -> bool:
    """
    Validate the prompt for the llama index model.
    The prompt must include the following placeholders:
    - {{text}}: The content string
    - {{num_questions}}: The number of questions to generate
    """
    if "{{text}}" not in prompt:
        raise ValueError("The prompt must include the placeholder {{text}}.")
    if "{{num_questions}}" not in prompt:
        raise ValueError("The prompt must include the placeholder {{num_questions}}.")
    return True


def parse_output(result: str) -> List[Dict]:
    result = result.strip()
    result = result.split("[Q]:")
    final_result = list()
    for res in result:
        res = res.strip()
        if res and "\n[A]:" in res:
            qa = res.split("\n[A]:")
            final_result.append({
                'query': qa[0].strip(),
                'generation_gt': qa[1].strip()
            })
    return final_result


def distribute_list_by_ratio(input_list, ratio) -> List[List[Any]]:
    total_ratio = sum(ratio)
    total_length = len(input_list)

    # Calculate the length of each slice
    slice_lengths = [int((r / total_ratio) * total_length) for r in ratio]

    # Adjust the last slice in case of rounding issues
    slice_lengths[-1] = total_length - sum(slice_lengths[:-1])

    slices = []
    start = 0
    for length in slice_lengths:
        end = start + length
        slices.append(input_list[start:end])
        start = end

    return slices



import os
import tempfile

import numpy as np
import pandas as pd
import pytest

from autorag import generator_models
from autorag.evaluate.util import cast_metrics
from autorag.nodes.generator import llama_index_llm
from autorag.nodes.promptmaker import fstring
from autorag.nodes.promptmaker.run import evaluate_generator_result, evaluate_one_prompt_maker_node, \
    run_prompt_maker_node
from autorag.utils.util import load_summary_file
from tests.mock import MockLLM

prompts = ['Hello, Do you know the world without war?',
           'Hi, I am dreaming about the world without any war.']
sample_generated_texts = ['hello. This is the world speaking',
                          'Hello, Do you know the world without war?']
sample_generation_gt = [
    ['Hello from space. Hi! This is the world speaking.'],
    ['Hello, Do you know the world without war?', 'Hi, I am dreaming about the world without any war.']
]
metrics = [{'metric_name': 'bleu'}, {'metric_name': 'rouge'}]
previous_result = pd.DataFrame({
    'query': ['What is the war?', 'Who is John Lennon?'],
    'retrieved_contents': [['War is horrible.', 'War is bad.'],
                           ['John Lennon is a musician.', 'John Lennon is a singer.']],
    'test_column': ['test_value_1', 'test_value_2'],
})


def test_evaluate_generator_result():
    sample_df = pd.DataFrame({'generated_texts': sample_generated_texts})
    result_df = evaluate_generator_result(sample_df, sample_generation_gt, metrics)
    metric_names, _ = cast_metrics(metrics)
    assert all(metric_name in result_df.columns for metric_name in metric_names)
    assert len(result_df) == len(sample_generated_texts)


def test_evaluate_one_prompt_maker_node():
    generator_models['mock'] = MockLLM
    generator_funcs = [llama_index_llm, llama_index_llm]
    generator_params = [{'llm': 'mock', 'model': 'gpt-3.5-turbo'},
                        {'llm': 'mock', 'model': 'gpt-3.5-turbo-1106'}]
    project_dir = '_'
    best_result = evaluate_one_prompt_maker_node(generator_funcs, generator_params, prompts, sample_generation_gt,
                                                 metrics, project_dir)
    metric_names, _ = cast_metrics(metrics)
    assert isinstance(best_result, pd.DataFrame)
    assert all(metric_name in best_result.columns for metric_name in metric_names)
    assert len(best_result) == len(prompts)


@pytest.fixture
def node_line_dir():
    with tempfile.TemporaryDirectory() as project_dir:
        data_dir = os.path.join(project_dir, "data")
        os.makedirs(data_dir)
        qa_data = pd.DataFrame({
            'qid': ['id-1', 'id-2'],
            'query': ['What is the war?', 'Who is John Lennon?'],
            'retrieval_gt': [[['doc-1']], [['doc-2']]],
            'generation_gt': sample_generation_gt,
        })
        qa_data.to_parquet(os.path.join(data_dir, "qa.parquet"), index=False)
        trial_dir = os.path.join(project_dir, "trial")
        os.makedirs(trial_dir)
        node_line_path = os.path.join(trial_dir, "node_line_1")
        os.makedirs(node_line_path)
        yield node_line_path


def check_best_result(best_df: pd.DataFrame):
    assert isinstance(best_df, pd.DataFrame)
    assert len(best_df) == len(previous_result)
    assert set(best_df.columns) == {
        'query', 'retrieved_contents', 'test_column', 'prompts', 'prompt_maker_bleu', 'prompt_maker_rouge'
    }


def check_summary_df(node_line_dir):
    # check the files saved properly
    summary_path = os.path.join(node_line_dir, "prompt_maker", "summary.csv")
    assert os.path.exists(summary_path)
    summary_df = load_summary_file(summary_path)
    assert len(summary_df) == len(previous_result)
    assert set(summary_df.columns) == {'filename', 'module_name', 'module_params', 'execution_time',
                                       'average_prompt_token', 'prompt_maker_bleu', 'prompt_maker_rouge', 'is_best'}
    best_filename = summary_df[summary_df['is_best']]['filename'].values[0]
    return best_filename


def test_run_prompt_maker_node(node_line_dir):
    generator_models['mock'] = MockLLM
    modules = [fstring, fstring]
    params = [{'prompt': 'Tell me something about the question: {query} \n\n {retrieved_contents}'},
              {'prompt': 'Question: {query} \n Something to read: {retrieved_contents} \n What\'s your answer?'}]
    strategies = {
        'metrics': metrics,
        'speed_threshold': 5,
        'token_threshold': 25,
        'tokenizer': 'gpt-3.5-turbo',
        'generator_modules': [{
            'module_type': 'llama_index_llm',
            'llm': 'mock',
            'model': ['gpt-3.5-turbo', 'gpt-3.5-turbo-1106'],
        }]
    }
    best_result = run_prompt_maker_node(modules, params, previous_result, node_line_dir, strategies)
    check_best_result(best_result)
    best_filename = check_summary_df(node_line_dir)
    assert best_filename == '0.parquet'
    best_result_path = os.path.join(node_line_dir, "prompt_maker", f"best_{best_filename}")
    assert os.path.exists(best_result_path)

    assert os.path.exists(os.path.join(node_line_dir, "prompt_maker", "0.parquet"))
    assert os.path.exists(os.path.join(node_line_dir, "prompt_maker", "1.parquet"))


def test_run_prompt_maker_node_default(node_line_dir):
    modules = [fstring, fstring]
    params = [{'prompt': 'Tell me something about the question: {query} \n\n {retrieved_contents}'},
              {'prompt': 'Question: {query} \n Something to read: {retrieved_contents} \n What\'s your answer?'}]
    strategies = {
        'metrics': metrics
    }
    best_result = run_prompt_maker_node(modules, params, previous_result, node_line_dir, strategies)
    check_best_result(best_result)
    best_filename = check_summary_df(node_line_dir)
    best_result_path = os.path.join(node_line_dir, "prompt_maker", f"best_{best_filename}")
    assert os.path.exists(best_result_path)


def test_run_prompt_maker_one_module(node_line_dir):
    modules = [fstring]
    params = [{'prompt': 'Tell me something about the question: {query} \n\n {retrieved_contents}'}]
    strategies = {
        'metrics': metrics
    }
    best_result = run_prompt_maker_node(modules, params, previous_result, node_line_dir, strategies)
    assert set(best_result.columns) == {
        'query', 'retrieved_contents', 'test_column', 'prompts'  # automatically skip evaluation
    }
    summary_filepath = os.path.join(node_line_dir, "prompt_maker", "summary.csv")
    assert os.path.exists(summary_filepath)
    summary_df = load_summary_file(summary_filepath)
    assert set(summary_df) == {
        'filename', 'module_name', 'module_params', 'execution_time', 'is_best', 'average_prompt_token',
    }
    best_filepath = os.path.join(node_line_dir, "prompt_maker", f"best_{summary_df['filename'].values[0]}")
    assert os.path.exists(best_filepath)


def test_summary_metric_name_at_threshold_cutoff():
    evaluation_results = [
        pd.DataFrame({
            'test_column': ['test_value_1', 'test_value_2', 'test_value_3'],
            'bleu': [0.1, 0.2, 0.3],
            'rouge': [0.4, 0.5, 0.6],
        })
    ]
    summary_df = pd.DataFrame({
        'filename': ['filename_1', 'filename_2'],
        'execution_time': [1, 2],
    })
    filenames = ['filename_2']
    metric_names, _ = cast_metrics(metrics)
    evaluation_df = pd.DataFrame({
        'filename': filenames,
        **{f'prompt_maker_{metric_name}': list(map(lambda x: x[metric_name].mean(), evaluation_results))
           for metric_name in metric_names}
    })
    summary_df = pd.merge(on='filename', left=summary_df, right=evaluation_df, how='left')

    assert set(summary_df.columns) == {'filename', 'execution_time', 'prompt_maker_bleu', 'prompt_maker_rouge'}
    assert np.isnan(summary_df['prompt_maker_bleu'].values[0])
    assert np.isnan(summary_df['prompt_maker_rouge'].values[0])
    assert summary_df['prompt_maker_bleu'].values[1] == pytest.approx(0.2)
    assert summary_df['prompt_maker_rouge'].values[1] == pytest.approx(0.5)



from typing import List

from autorag.nodes.promptmaker.base import prompt_maker_node


@prompt_maker_node
def fstring(prompt: str,
            queries: List[str], retrieved_contents: List[List[str]]) -> List[str]:
    """
    Make a prompt using f-string from a query and retrieved_contents.
    You must type a prompt or prompt list at config yaml file like this:

    .. Code:: yaml
    nodes:
    - node_type: prompt_maker
      modules:
      - module_type: fstring
        prompt: [Answer this question: {query} \n\n {retrieved_contents},
        Read the passages carefully and answer this question: {query} \n\n Passages: {retrieved_contents}]

    :param prompt: A prompt string.
    :param queries: List of query strings.
    :param retrieved_contents: List of retrieved contents.
    :return: Prompts that made by f-string.
    """
    def fstring_row(_prompt: str, _query: str, _retrieved_contents: List[str]) -> str:
        contents_str = "\n\n".join(_retrieved_contents)
        return _prompt.format(query=_query, retrieved_contents=contents_str)

    return list(map(lambda x: fstring_row(prompt, x[0], x[1]), zip(queries, retrieved_contents)))



import asyncio
from typing import List, Tuple

import torch
from transformers import T5Tokenizer, T5ForConditionalGeneration

from autorag.nodes.passagereranker.base import passage_reranker_node

prediction_tokens = {
    'castorini/monot5-base-msmarco': ['▁false', '▁true'],
    'castorini/monot5-base-msmarco-10k': ['▁false', '▁true'],
    'castorini/monot5-large-msmarco': ['▁false', '▁true'],
    'castorini/monot5-large-msmarco-10k': ['▁false', '▁true'],
    'castorini/monot5-base-med-msmarco': ['▁false', '▁true'],
    'castorini/monot5-3b-med-msmarco': ['▁false', '▁true'],
    'castorini/monot5-3b-msmarco-10k': ['▁false', '▁true'],
    'unicamp-dl/mt5-base-en-msmarco': ['▁no', '▁yes'],
    'unicamp-dl/ptt5-base-pt-msmarco-10k-v2': ['▁não', '▁sim'],
    'unicamp-dl/ptt5-base-pt-msmarco-100k-v2': ['▁não', '▁sim'],
    'unicamp-dl/ptt5-base-en-pt-msmarco-100k-v2': ['▁não', '▁sim'],
    'unicamp-dl/mt5-base-en-pt-msmarco-v2': ['▁no', '▁yes'],
    'unicamp-dl/mt5-base-mmarco-v2': ['▁no', '▁yes'],
    'unicamp-dl/mt5-base-en-pt-msmarco-v1': ['▁no', '▁yes'],
    'unicamp-dl/mt5-base-mmarco-v1': ['▁no', '▁yes'],
    'unicamp-dl/ptt5-base-pt-msmarco-10k-v1': ['▁não', '▁sim'],
    'unicamp-dl/ptt5-base-pt-msmarco-100k-v1': ['▁não', '▁sim'],
    'unicamp-dl/ptt5-base-en-pt-msmarco-10k-v1': ['▁não', '▁sim'],
    'unicamp-dl/mt5-3B-mmarco-en-pt': ['▁', '▁true'],
    'unicamp-dl/mt5-13b-mmarco-100k': ['▁', '▁true'],
}


@passage_reranker_node
def monot5(queries: List[str], contents_list: List[List[str]],
           scores_list: List[List[float]], ids_list: List[List[str]],
           top_k: int, model_name: str = 'castorini/monot5-3b-msmarco-10k') \
        -> Tuple[List[List[str]], List[List[str]], List[List[float]]]:
    """
    Rerank a list of contents based on their relevance to a query using MonoT5.

    :param queries: The list of queries to use for reranking
    :param contents_list: The list of lists of contents to rerank
    :param scores_list: The list of lists of scores retrieved from the initial ranking
    :param ids_list: The list of lists of ids retrieved from the initial ranking
    :param top_k: The number of passages to be retrieved
    :param model_name: The name of the MonoT5 model to use for reranking
        Note: default model name is 'castorini/monot5-3b-msmarco-10k'
            If there is a '/' in the model name parameter,
            when we create the file to store the results, the path will be twisted because of the '/'.
            Therefore, it will be received as '_' instead of '/'.
    :return: tuple of lists containing the reranked contents, ids, and scores
    """
    # replace '_' to '/'
    if '_' in model_name:
        model_name = model_name.replace('_', '/')
    # Load the tokenizer and model from the pre-trained MonoT5 model
    tokenizer = T5Tokenizer.from_pretrained(model_name)
    model = T5ForConditionalGeneration.from_pretrained(model_name).eval()
    # Retrieve the tokens used by the model to represent false and true predictions
    token_false, token_true = prediction_tokens[model_name]
    token_false_id = tokenizer.convert_tokens_to_ids(token_false)
    token_true_id = tokenizer.convert_tokens_to_ids(token_true)
    # Determine the device to run the model on (GPU if available, otherwise CPU)
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model.to(device)
    # Run async mono_t5_rerank_pure function
    tasks = [mono_t5_pure(query, contents, scores, top_k, ids, model, device, tokenizer, token_false_id, token_true_id) \
             for query, contents, scores, ids in zip(queries, contents_list, scores_list, ids_list)]
    loop = asyncio.get_event_loop()
    results = loop.run_until_complete(asyncio.gather(*tasks))
    content_result = list(map(lambda x: x[0], results))
    id_result = list(map(lambda x: x[1], results))
    score_result = list(map(lambda x: x[2], results))

    del model
    del tokenizer
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    return content_result, id_result, score_result


async def mono_t5_pure(query: str, contents: List[str], scores: List[float], top_k: int,
                       ids: List[str], model, device, tokenizer, token_false_id, token_true_id)\
        -> Tuple[List[str], List[str], List[float]]:
    """
    Rerank a list of contents based on their relevance to a query using MonoT5.

    :param query: The query to use for reranking
    :param contents: The list of contents to rerank
    :param scores: The list of scores retrieved from the initial ranking
    :param ids: The list of ids retrieved from the initial ranking
    :param model: The MonoT5 model to use for reranking
    :param device: The device to run the model on (GPU if available, otherwise CPU)
    :param tokenizer: The tokenizer to use for the model
    :param token_false_id: The id of the token used by the model to represent a false prediction
    :param token_true_id: The id of the token used by the model to represent a true prediction
    :return: tuple of lists containing the reranked contents, ids, and scores
    """

    # Format the input for the model by combining each content with the query
    input_texts = [f'Query: {query} Document: {content}' for content in contents]
    # Tokenize the input texts and prepare for model input
    input_encodings = tokenizer(input_texts, padding=True, truncation=True, max_length=512, return_tensors='pt').to(
        device)

    # Generate model predictions without updating model weights
    with torch.no_grad():
        outputs = model.generate(input_ids=input_encodings['input_ids'],
                                 attention_mask=input_encodings['attention_mask'],
                                 output_scores=True,
                                 return_dict_in_generate=True)

    # Extract logits for the 'false' and 'true' tokens from the model's output
    logits = outputs.scores[-1][:, [token_false_id, token_true_id]]
    # Calculate the softmax probability of the 'true' token
    probs = torch.nn.functional.softmax(logits, dim=-1)[:, 1]  # Get the probability of the 'true' token

    # Create a list of tuples pairing each content with its relevance probability
    content_ids_probs = list(zip(contents, ids, probs.tolist()))

    # Sort the list of pairs based on the relevance score in descending order
    sorted_content_ids_probs = sorted(content_ids_probs, key=lambda x: x[2], reverse=True)

    # crop with top_k
    if len(contents) < top_k:
        top_k = len(contents)
    sorted_content_ids_probs = sorted_content_ids_probs[:top_k]

    content_result, id_result, score_result = zip(*sorted_content_ids_probs)

    return list(content_result), list(id_result), list(score_result)



from typing import List

import pandas as pd

from autorag import generator_models
from autorag.nodes.passagecompressor import tree_summarize
from tests.mock import MockLLM

queries = [
    "What is the capital of France?",
    "What is the meaning of life?",
]
retrieved_contents = [
    ["Paris is the capital of France.", "France is a country in Europe.", "France is a member of the EU."],
    ["The meaning of life is 42.", "The meaning of life is to be happy.", "The meaning of life is to be kind."],
]


def check_result(result: List[str]):
    assert len(result) == len(queries)
    for r in result:
        assert isinstance(r, str)
        assert len(r) > 0
        assert bool(r) is True


def test_tree_summarize_default():
    llm = MockLLM()
    result = tree_summarize.__wrapped__(queries, retrieved_contents, [], [], llm)
    check_result(result)


def test_tree_summarize_chat():
    gpt_3 = MockLLM(model="gpt-3.5-turbo")
    result = tree_summarize.__wrapped__(queries, retrieved_contents, [], [], gpt_3)
    check_result(result)


def test_tree_summarize_custom_prompt():
    llm = MockLLM()
    prompt = "This is a custom prompt. {context_str} {query_str}"
    result = tree_summarize.__wrapped__(queries, retrieved_contents, [], [], llm, prompt=prompt)
    check_result(result)
    assert bool(result[0]) is True


def test_tree_summarize_custom_prompt_chat():
    gpt_3 = MockLLM(model="gpt-3.5-turbo")
    prompt = "Query: {query_str} Passages: {context_str}. Repeat the query."
    result = tree_summarize.__wrapped__(queries, retrieved_contents, [], [], gpt_3, chat_prompt=prompt)
    check_result(result)
    assert 'What is the capital of France?' in result[0]


def test_tree_summarize_node():
    generator_models['mock'] = MockLLM
    df = pd.DataFrame({
        'query': queries,
        'retrieved_contents': retrieved_contents,
        'retrieved_ids': [['id-1', 'id-2', 'id-3'], ['id-4', 'id-5', 'id-6']],
        'retrieve_scores': [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]],
    })
    result = tree_summarize(
        "project_dir",
        df,
        llm='mock',
        prompt="This is a custom prompt. {context_str} {query_str}",
        max_tokens=64,
    )
    assert isinstance(result, pd.DataFrame)
    contents = result['retrieved_contents'].tolist()
    assert isinstance(contents, list)
    assert len(contents) == len(queries)
    assert isinstance(contents[0], list)
    assert len(contents[0]) == 1
    assert isinstance(contents[0][0], str)
    assert bool(contents[0][0]) is True
